;; Scalable Sawfish theme resembling Mac OS 7.5.3 window decorations.
;; Copyright (c) 2001  Andrew Chadwick
;; 
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of the
;; License, or (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software Foundation,
;; Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

(require 'gtkrc)			; for colour calculations


;;;; Shape rasterisation


(define (render-filled-box-% img %x1 %y1 %x2 %y2 col . opt)
  "Renders a filled box into an image using percentage measurements
and optional pixel offsets. Units for %x1, %y1, %x2, %y2, are in
percentages of img's width and height.

The optional arguments, if supplied, are (dx1 dy1 dx2 dy2 alpha). dx1,
dy1 etc. are pixel offsets, and alpha is the amount of col to use when
blending, from 0 to 1. Pixel offsets default to zero, and alpha
defaults to one."
  (let* ((w (car (image-dimensions img)))
	 (h (cdr (image-dimensions img)))
	 (dx1 (or (nth 0 opt) 0))	; Handle options
	 (dy1 (or (nth 1 opt) 0))
	 (dx2 (or (nth 2 opt) 0))
	 (dy2 (or (nth 3 opt) 0))
	 (alpha (or (nth 4 opt) 1.0))
	 (x1 (/ (* %x1 w) 100))		; First, scale
	 (y1 (/ (* %y1 h) 100))
	 (x2 (/ (* %x2 w) 100))
	 (y2 (/ (* %y2 h) 100))
	 (x1 (max 0 (min (- w 1) x1)))	; Then clamp to the image size
	 (y1 (max 0 (min (- h 1) y1)))
	 (x2 (max 0 (min (- w 1) x2)))
	 (y2 (max 0 (min (- h 1) y2)))
	 (x1 (+ x1 dx1))		; Finally, apply pixel offsets
	 (y1 (+ y1 dy1))
	 (x2 (+ x2 dx2))
	 (y2 (+ y2 dy2)))
    (render-filled-box img x1 y1 x2 y2 col alpha)))


(define (render-filled-box image x1 y1 x2 y2 col . opt)
  "Renders a filled box into an image using pixel measurements only.
For a proportional approach, use (render-filled-box-%) instead. The
optional argument, if supplied, is alpha, the amount of col to use
when blending. Alpha is expressed as a floating point number ranging
from 0.0 to 1.0. The default is 1.0."
  (let* ((alpha (or (car opt) 1.0))
	 (w (car (image-dimensions image)))
	 (h (cdr (image-dimensions image)))
	 (x (max 0 x1)))
    (while (and (<= x x2) (< x w))
	   (let* ((y (max 0 y1)))
	     (while (and (<= y y2) (< y h))
		    (let* ((cur (rgba-list->color (image-ref image x y)))
			   (mix (mix-colours col cur alpha)))
		      (image-set image x y (color->rgba-list mix)))
		    (setq y (+ 1 y))))
	   (setq x (+ 1 x)))))


(define (render-line-% img %x1 %y1 %x2 %y2 col . opt)
  "Renders a single-pixel-wide into an image using percentage
measurements and optional pixel offsets. Units for %x1, %y1, %x2, %y2,
are in percentages of img's width and height.

The optional arguments, if supplied, are (dx1 dy1 dx2 dy2 alpha),
where dx1, dy1 etc. are pixel offsets, and alpha is the amount of col
to use when blending, from 0 to 1. Pixel offsets default to zero, and
alpha defaults to one."
  (let* ((w (car (image-dimensions img)))
	 (h (cdr (image-dimensions img)))
	 (dx1 (or (nth 0 opt) 0))	; Handle options
	 (dy1 (or (nth 1 opt) 0))
	 (dx2 (or (nth 2 opt) 0))
	 (dy2 (or (nth 3 opt) 0))
	 (alpha (or (nth 4 opt) 1.0))
	 (x1 (/ (* %x1 w) 100))		; First, scale
	 (y1 (/ (* %y1 h) 100))
	 (x2 (/ (* %x2 w) 100))
	 (y2 (/ (* %y2 h) 100))
	 (x1 (max 0 (min (- w 1) x1)))	; Then clamp to the image size
	 (y1 (max 0 (min (- h 1) y1)))
	 (x2 (max 0 (min (- w 1) x2)))
	 (y2 (max 0 (min (- h 1) y2)))
	 (x1 (+ x1 dx1))		; Finally, apply pixel offsets
	 (y1 (+ y1 dy1))
	 (x2 (+ x2 dx2))
	 (y2 (+ y2 dy2)))
    (render-line img x1 y1 x2 y2 col alpha)))


(define (render-line image x y x2 y2 col . opt)
  "Renders a single-pixel-wide line into an image using pixel
measurements. For a proportional approach, use (render-line-%)
instead. The optional argument, if supplied, is alpha, the amount of
col to use when blending. Alpha is expressed as a floating point
number ranging from 0.0 to 1.0. The default is 1.0"
  (let* ((alpha (or (car opt) 1.0))
	 (w (car (image-dimensions image)))
	 (h (cdr (image-dimensions image)))
	 (dx (- x2 x))  (dy (- y2 y))
	 (long-d 0)     (short-d 0)
	 (d 0)      (add-dh 0)  (add-dl 0)
	 (inc-xh 0) (inc-yh 0)  (inc-xl 0) (inc-yl 0)
	 (i 0))
    (if (< dx 0)
	(setq dx (- dx) inc-xh -1 inc-xl -1)
	(setq inc-xh 1 inc-xl 1))
    (if (< dy 0)
	(setq dy (- dy) inc-yh -1 inc-yl -1)
	(setq inc-yh 1  inc-yl 1))
    (if (> dx dy)
	(setq long-d dx short-d dy inc-yl 0)
	(setq long-d dy short-d dx inc-xl 0))
    (setq d      (* 2 (- short-d long-d))
	  add-dl (* 2 short-d)
	  add-dh (- (* 2 short-d) (* 2 long-d)))    
    (while (<= i long-d)
	   (if (and (>= x 0) (>= y 0)
		    (< x w) (< y h))
	       (let* ((cur (rgba-list->color (image-ref image x y)))
		      (mix (mix-colours col cur alpha)))
		 (image-set image x y (color->rgba-list mix))) )
	   (if (>= d 0)
	       (setq x (+ x inc-xh) y (+ y inc-yh) d (+ d add-dh))
	       (setq x (+ x inc-xl) y (+ y inc-yl) d (+ d add-dl)))
	   (setq i (+ 1 i)))
    image))



;;;; Colour mixing and RGBA lists


(define (mix-colours col1 col2 p1)
  "Mixes two colours together, returning a new colour. col1 and col2
are (colorp) objects, p1 is a number in the range 0.0 to 1.0 giving
the proportion of col1 to use in the returned value."
  (define (.mix-els e1 e2)
    (inexact->exact
     (round (* 255 (+ (* e1 p1) (* e2 (- 1.0 p1)))))))
  (cond ((>= p1 1.0) col1)
	((<= p1 0.0) col2)
	(t (let* ((hi (color-rgb-8 col1))
		  (lo (color-rgb-8 col2)))
	     (get-color-rgb (.mix-els (car hi) (car lo))
			    (.mix-els (cadr hi) (cadr lo))
			    (.mix-els (caddr hi) (caddr lo)))) )))

(define (color->rgba-list col)
  "Converts a Sawfish colour object into an RGBA list. The returned
list has an alpha value of 255."
  (nconc (color-rgb-8 col) '(255)))


(define (rgba-list->color col-ls)
  "Converts an RGBA list representing a colour into a Sawfish colour
object. The alpha value of the list is ignored."
  (get-color-rgb (car col-ls) (cadr col-ls) (caddr col-ls)))



;; The following functions and definitions are specific to the
;; HappyDesktop theme only.


;;;; Customisation options


(defgroup HappyDesktop "HappyDesktop theme"
  :group appearance)

(defcustom HappyDesktop:font default-font
  "Titlebar font."
  :type font
  :group (appearance HappyDesktop)
  :after-set (lambda (var)
	       (rebuild-images)
	       (rebuild-frames)
	       (after-setting-frame-option)))

(defcustom HappyDesktop:colour-scheme 'Default
  "Accent colour."
  :group (appearance HappyDesktop)
  :type (set Default Gold Green Cyan Red Magenta Blue Grey) ;; TODO: add 'B&W
  :after-set (lambda nil
	       (rebuild-colours)
	       (rebuild-images)
	       (rebuild-frames)
	       (after-setting-frame-option)))

(defgroup Authenticity "Authenticity"
  :group (appearance HappyDesktop))

(defcustom HappyDesktop:mac-like-resize nil
  "Corner resize tab."
  :type boolean
  :group (appearance HappyDesktop Authenticity)
  :after-set (lambda (var)
	       (rebuild-images)
	       (rebuild-frames)
	       (after-setting-frame-option)))

(defcustom HappyDesktop:shadows t
  "Window shadows."
  :type boolean
  :group (appearance HappyDesktop Authenticity)
  :after-set (lambda (var)
	       (rebuild-frames)
	       (after-setting-frame-option)))


;; FIXME: implement this for 1.0:
;;(defcustom HappyDesktop:transients-always-focused nil
;;  "Transients always appear focused."
;;  :type boolean
;;  :group (appearance HappyDesktop Authenticity)
;;  :after-set (lambda (var)
;;	       (rebuild-images)
;;	       (after-setting-frame-option)))



;;;; Window dimension calculations

(define minimum-stipple-height 7)	;Doesn't include rounding up.

(define (get-vertical-padding)
  (let ((fh (font-height HappyDesktop:font)))
    (/ (max (/ fh 3) 8) 2)))

(define (get-horizontal-padding)
  (max 1 (/ (get-vertical-padding) 3)))

(define (get-stipple-height)
  (let* ((font-ht (font-height HappyDesktop:font))
	 (tbar-ht-min (+ font-ht 4))
	 (vpad (get-vertical-padding))
	 (stip-ht-min (max (- tbar-ht-min (* 2 vpad))
			   minimum-stipple-height)))
    (+ stip-ht-min -1 (- 2 (% stip-ht-min 2))))) ;Round to next (1+2n).

(define (get-transient-stipple-height)
  (let* ((fh (font-height HappyDesktop:font))
	 (sh (max (/ (* 3 fh) 5)
		  minimum-stipple-height))) ;Transients don't have text, so
    (+ sh -2 (- 3 (% sh 3)))))		;no need to bother with padding.

(define (get-titlebar-height)
  (+ (get-stipple-height)
     (* 2 (get-vertical-padding))))

(define (get-transient-titlebar-height)
  (+ (get-transient-stipple-height)
     (* 2 (get-vertical-padding))))


;;;; Window dimension thunks

;; Some window dimensions can only be calculated on the fly because
;; they depend on X properties of the window. Try to keep these to a
;; minimum to keep redraws fast.


(define (get-label-width w)
  "Gets the width of the label. This is only valid for standard
windows, but then they're the only ones that are labelled."
  (let* ((button-wd (car (image-dimensions (car close-button-images))))
	 (button-space (+ button-wd (/ button-wd 2))))
    (min (- (car (window-dimensions w)) button-space)
	 (+ (text-width (window-name w) HappyDesktop:font)
	    (font-height HappyDesktop:font)))))
	 
(define (get-label-left-edge w)
  (let* ((font-ht (font-height HappyDesktop:font)))
    (/ (- (car (window-dimensions w))
	  (get-label-width w))
       2)))



;;;; Colours

(define colour-schemes
  `((Default . (,(get-color "#333366") . ,(get-color "#ccccff")))
    (Gold    . (,(get-color "#996600") . ,(get-color "#ffcc99")))
    (Green   . (,(get-color "#003300") . ,(get-color "#99cc99")))
    (Cyan    . (,(get-color "#003333") . ,(get-color "#99ffff")))
    (Red     . (,(get-color "#990000") . ,(get-color "#ffcccc")))
    (Magenta . (,(get-color "#990066") . ,(get-color "#ffccff")))
    (Blue    . (,(get-color "#003366") . ,(get-color "#99ccff")))
    (Grey    . (,(get-color "#333333") . ,(get-color "#dddddd")))
    (B&W     . (,(get-color "#ffffff") . ,(get-color "#ffffff"))) )) ;TODO

(define col.borders          (get-color "#000000"))
(define col.title-text       (get-color "#000000"))
(define col.title-bg         (get-color "#eeeeee"))
(define col.title-bg-details (get-color "#777777"))
(define col.button-bg        (get-color "#a4a4a4"))
(define col.button-bg-hi     (get-color "#bbbbbb"))

(define col.borders-desel    nil)
(define col.accent-hi        nil)
(define col.accent-mid       nil)
(define col.accent-lo        nil)
(define col.title-bg-desel   nil)
(define col.title-text-desel nil)
(define col.active-button         nil)
(define col.active-button-details nil)

(define (rebuild-colours)
  "Recalculates the colours used by rebuild-frames and rebuild-images.
Call whenever the colour scheme changes."
  (setq col.accent-hi (cdr (cdr (assoc HappyDesktop:colour-scheme
				       colour-schemes)))
	col.accent-lo (car (cdr (assoc HappyDesktop:colour-scheme
				       colour-schemes)))
	col.accent-mid (mix-colours col.accent-hi col.accent-lo 0.75)
	col.title-bg-desel (mix-colours (car gtkrc-background)
					(car gtkrc-base)
					0.75)
	col.title-text-desel (mix-colours
			      col.title-bg-desel col.title-text 0.5)
	col.borders-desel (mix-colours col.borders col.title-bg 0.75)
	col.active-button (mix-colours col.accent-hi col.button-bg 0.5)
	col.active-button-details (mix-colours
				   col.borders col.accent-mid 0.66) ))


;;;; Button image factories

(define (make-unfocused-button h)
  (let* ((pad (get-horizontal-padding))
	 (w (+ (* 2 pad) h))
	 (img (make-sized-image w h)))
    (clear-image img col.title-bg-desel)))

(define (make-plain-button h)
  (let* ((pad (get-horizontal-padding))
	 (-pad (- pad))
	 (w (+ (* 2 pad) h))
	 (img (make-sized-image w h)))
    (clear-image img col.title-bg)
    (render-filled-box-% img 0 0 100 100 col.accent-lo pad 0 -pad 0)
    (render-filled-box-% img 0 0 100 100 col.accent-hi
			 (+ pad 1) 1 -pad 0)
    (render-filled-box-% img 0 0 100 100 col.accent-lo
			 (+ pad 2) 2 (- -pad 1) -1)
    (render-filled-box-% img 0 0 100 100 col.button-bg
			 (+ pad 2) 2 (- -pad 2) -2)
    img))

(define (make-maximise-button h)
  (let* ((img (make-plain-button h))
	 (pad (get-horizontal-padding))
	 (w (car (image-dimensions img)))
	 (h (cdr (image-dimensions img)))
	 (off (/ h 2))
	 (-pad (- pad)))
    (when (> h 7)			;Image is too complex otherwise
	  (render-filled-box-% img 0 0 100 100 col.accent-lo
			       (+ pad 2) 2 (- 1 (+ pad off)) (- 1 off))
	  (render-filled-box-% img 0 0 100 100 col.button-bg
			       (+ pad 2) 2 (- (+ pad off)) (- off)))
    img))

(define (make-active-button h)
  (let* ((pad (get-horizontal-padding))
	 (-pad (- pad))
	 (w (+ (* 2 pad) h))
	 (img (make-sized-image w h)))
    (clear-image img col.title-bg)
    ;; Box and outline
    (render-filled-box-% img 0 0 100 100 col.active-button-details
			 pad 0 -pad 0)
    (render-filled-box-% img 0 0 100 100 col.active-button
			 (+ pad 1) 1 (+ -1 -pad) -1)
    ;; Diagonals
    (when (> h 7)
	  (render-line-% img 0 0 100 100 col.active-button-details
			 (+ (* 2 pad) 1) (+ 1 pad)
			 (- (* 2 -pad) 1) (+ -1 -pad))
	  (render-line-% img 0 100 100 0 col.active-button-details
			 (+ (* 2 pad) 1) (+ -1 -pad)
			 (- (* 2 -pad) 1) (+ 1 pad)))
    ;; Horizontals
    (render-line-% img 0 50 100 50 col.active-button-details
		   (* 2 pad) 0 (* 2 -pad) 0)
    (render-line-% img 50 0 50 100 col.active-button-details
		   0 1 0 -1)
    ;; Centre spot
    (render-filled-box-% img 50 50 50 50 col.active-button
			 -pad -pad pad pad)
    img))


;;;; Resize tab/bar image factories

(define (make-mac-resize-tab focused siz)
  (let* ((img (make-sized-image siz siz)))
    (cond (focused
	   (clear-image img col.borders)
	   (render-filled-box-% img 0 0 100 100 col.title-bg 1 1 0 0)
	   ;; Lower portion:
	   (render-filled-box-% img 28 28 86 86 col.accent-lo 1 1 0 0)
	   (render-filled-box-% img 28 28 86 86 col.button-bg 2 2 -1 -1)
	   (render-filled-box-% img 28 28 86 28 col.accent-hi 2 2 0 2)
	   (render-filled-box-% img 28 28 28 86 col.accent-hi 2 2 2 0)
	   ;; Upper portion
	   (render-filled-box-% img 14 14 56 56 col.accent-lo 1 1 0 0)
	   (render-filled-box-% img 14 14 56 56 col.button-bg-hi 2 2 -1 -1)
	   (render-filled-box-% img 14 14 56 14 col.accent-hi 2 2 0 2)
	   (render-filled-box-% img 14 14 14 56 col.accent-hi 2 2 2 0))
	  (t
	   (clear-image img col.borders-desel)
	   (render-filled-box-% img 0 0 100 100 col.title-bg-desel 1 1 0 0)))
    (image-put img 'tiled t)
    img))

;; Image factories for "X-style" (well, neXt-style) resize bars

(define (make-x-resize-edge focused)
  (let* ((img (make-sized-image 5 5)))
    (cond (focused
	   (clear-image img col.title-bg)
	   (render-focused-bevels img))
	  (t
	   (clear-image img col.title-bg-desel)
	   (render-unfocused-bevels img)))
    (image-put img 'tiled nil)
    img))

(define (make-x-resize-grip focused side font-ht)
  (let* ((ht0 (max 4 (/ font-ht 3)))
	 (unit-x (+ 1 ht0))
	 (width (* 5 unit-x))
	 (height (+ 2 ht0))
	 (img (make-sized-image width height))
	 (top 1)
	 (bottom (- height 2))
	 (inner (- bottom top))
	 (bg-col (if focused col.title-bg col.title-bg-desel)))
    ;; Create the backdrop
    (clear-image img (if focused col.borders col.borders-desel))
    (render-filled-box img 0 top width bottom bg-col)
    (when focused
	  ;; Accent lines top and bottom
	  (render-line img 0 top width top col.accent-hi)
	  (render-line img 0 bottom width bottom col.accent-mid)
	  ;; Render grips
	  (let* ((xmax (* 4 unit-x))
		 (xmin unit-x)
		 (x xmin)
		 (t (if (eq side 'left) bottom top))
		 (b (if (eq side 'left) top bottom))
		 (-b1 (if (eq side 'left) (+ top 1) (- bottom 1))))
	    (while (< x xmax)
		   ;; (render-line img x b (+ x inner) t col.accent-mid)
		   (render-line img (+ 1 x) b (+ 1 x inner) t
				col.title-bg-details)
		   (render-line img (+ 3 x) -b1 (+ 2 x inner) t
				col.accent-hi)
		   (setq x (+ x unit-x))))
	  ;; Finishing touches
	  (let ((lx (- width 1)))
	    (cond ((eq side 'left)
		   (render-line img 0 top 0 (- bottom 1) col.accent-hi))
		  (t
		   (render-line img lx top lx (- bottom 1) col.accent-mid)))))
    (image-put img 'tiled t)
    img))


;;;; Other image factories

(define (render-focused-bevels img)
  "Renders an accented bevel into an image."
  (when (imagep img)
	;; Borders
	(render-filled-box-% img 0 0 0 100 col.borders     0 0 0 0)
	(render-filled-box-% img 100 0 100 100 col.borders 0 0 0 0)
	(render-filled-box-% img 0 0 100 0 col.borders     0 0 0 0)
	(render-filled-box-% img 0 100 100 100 col.borders 0 0 0 0)
	;; Accents
	(render-filled-box-% img 0 0 0 100 col.accent-hi 1 1 1 -2)
	(render-filled-box-% img 0 0 100 0 col.accent-hi 1 1 -2 1)
	(render-filled-box-% img 0 100 100 100 col.accent-mid 1 -1 -1 -1)
	(render-filled-box-% img 100 0 100 100 col.accent-mid -1 1 -1 -1)
	;; Set the border
	(set-image-border img 2 2 2 2))
  img)

(define (render-unfocused-bevels img)
  (when (imagep img)
	;; Borders
	(render-filled-box-% img 0 0 0 100 col.borders-desel 0 0 0 0)
	(render-filled-box-% img 100 0 100 100 col.borders-desel   0 0 0 0)
	(render-filled-box-% img 0 0 100 0 col.borders-desel 0 0 0 0)
	(render-filled-box-% img 0 100 100 100 col.borders-desel   0 0 0 0)
	(set-image-border img 1 1 1 1))
  img)

(define (make-stipple focused)
  "Makes the pattern which is tiled on top of ordinary titlebars."
  (let* ((img (make-sized-image 16 16))
	 (detail-col (color->rgba-list col.title-bg-details))
	 (bg-col (color->rgba-list col.title-bg)))
    (clear-image img col.title-bg-desel)
    (if focused
	(image-fill (lambda (x y)
		      (if (= (% y 2) 0)
			  detail-col
			  bg-col))
		    img))
    (image-put img 'tiled t)
    img))

(define (make-transient-stipple focused)
  "Makes the pattern which is tiled on top of transient titlebars."
  (let* ((img (make-sized-image 15 15))
	 (detail-col (color->rgba-list col.title-bg-details))
	 (bg-col (color->rgba-list col.title-bg)))
    (clear-image img col.title-bg-desel)
    (if focused
	(image-fill (lambda (x y)
		      (if (and (= (% x 3) 1)
			       (= (% y 3) 0))
			  detail-col
			  bg-col))
		    img))
    (image-put img 'tiled t)
    img))

(define (make-titlebar-background focused)
  (let* ((img (make-sized-image 5 5)))
    (if focused
	(progn (clear-image img col.title-bg)
	       (render-focused-bevels img))
	(progn (clear-image img col.title-bg-desel)
		(render-unfocused-bevels img)))
    img))


;;;; Button and pattern image placeholders

(define titlebar-background-images nil)
(define stipple-images             nil)
(define close-button-images        nil)
(define maximise-button-images     nil)

(define transient-titlebar-background-images nil)
(define transient-stipple-images             nil)
(define transient-close-button-images        nil)
(define transient-maximise-button-images     nil)

(define bottom-resize-edge-images  nil)
(define left-resize-corner-images  nil)
(define right-resize-corner-images nil)

(define (rebuild-images)
  "Rebuilds every image which must be created on the fly. Called on
startup and when customisable variables change."
  (let* ((font-ht (font-height HappyDesktop:font)) 
	 (stipple-ht (get-stipple-height))
	 (tr-stipple-ht (get-transient-stipple-height)))
    (setq
     titlebar-background-images (list (make-titlebar-background nil)
				      (make-titlebar-background t))
     stipple-images (list (make-stipple nil)
			  (make-stipple t))
     transient-titlebar-background-images titlebar-background-images
     transient-stipple-images (list (make-transient-stipple nil)
				    (make-transient-stipple t)))
    (if HappyDesktop:mac-like-resize
	(setq bottom-resize-edge-images nil
	      left-resize-corner-images nil
	      right-resize-corner-images (list (make-mac-resize-tab
						 nil font-ht)
					       (make-mac-resize-tab
						t font-ht)))
	(setq bottom-resize-edge-images (list (make-x-resize-edge nil)
					      (make-x-resize-edge t))
	      left-resize-corner-images (list (make-x-resize-grip
					       nil 'left font-ht)
					      (make-x-resize-grip
					       t 'left font-ht))
	      right-resize-corner-images (list (make-x-resize-grip
						nil 'right font-ht)
					       (make-x-resize-grip
						t 'right font-ht))))
    
    (let* ((plain-button (make-plain-button stipple-ht))
	   (maximise-button (make-maximise-button stipple-ht))
	   (unfocused-button (make-unfocused-button stipple-ht))
	   (active-button (make-active-button stipple-ht)))
      (setq close-button-images (list unfocused-button
				      plain-button
				      plain-button
				      active-button)
	    maximise-button-images (list unfocused-button
					 maximise-button
					 maximise-button
					 active-button) ))
    (let* ((tr-plain-button (make-plain-button tr-stipple-ht))
	   (tr-maximise-button (make-maximise-button tr-stipple-ht))
	   (tr-unfocused-button (make-unfocused-button tr-stipple-ht))
	   (tr-active-button (make-active-button tr-stipple-ht)))
      (setq transient-close-button-images
	    (list tr-unfocused-button
		  tr-plain-button
		  tr-plain-button
		  tr-active-button)
	    transient-maximise-button-images
	    (list tr-unfocused-button
		  tr-maximise-button
		  tr-maximise-button
		  tr-active-button) )) ))



;; Frame placeholders and builder

(define standard-frame         nil)
(define shaded-standard-frame  nil)
(define transient-frame        nil)
(define shaded-transient-frame nil)

(define (fixed-size-p w)
  (memq 'maximize-button
	(window-get w 'removed-classes)))

(define (maximisable-p w)
  (not (fixed-size-p w)))

(define (get-dim-if-maximisable w imgs height-p)
  "Gets the height or width of an image (or the first of a sequence of
images), unless 'w' doesn't appear to be resizable, in which case
return zero. Needs a better name."
  (let* ((img (if (consp imgs) (car imgs) imgs))
	 (func (if height-p cdr car)))
    (if (imagep img)
	(if (maximisable-p w)
	    (func (image-dimensions img))
	    0)
	0)))

(define (rebuild-frames)
  "Rebuilds the frame definitions used in this theme, 'fra.*', without
rebuilding any images. See 'rebuild-images'."
  (let* (;; Shared dimensions
	 (font-ht (font-height HappyDesktop:font))
	 (vpad (get-vertical-padding))
	 (hpad (get-horizontal-padding))

	 ;; Standard titlebar dims
	 (stipple-top (- vpad (get-titlebar-height)))
	 (button-ht (cdr (image-dimensions (car close-button-images))))
	 (button-wd (car (image-dimensions (car close-button-images))))

	 ;; Transient titlebar bits
	 (tr-stipple-top (- vpad (get-transient-titlebar-height)))
	 (tr-button-wd (car (image-dimensions
			     (car transient-close-button-images))))
	 (tr-button-ht (cdr (image-dimensions
			     (car transient-close-button-images))))
	 
	 ;; Resize control dimensions
	 (resize-corner-wd (car (image-dimensions
				 (car right-resize-corner-images) )))
	 (resize-corner-ht (cdr (image-dimensions
				 (car right-resize-corner-images) )))
	 (resize-edge-ht resize-corner-ht)
	 
	 ;; Build parts of the frame definitions piecemeal, combine in
	 ;; the body into the real thing.
	 (titlebar
	  `(((class      . title)
	     (left-edge  . -1)
	     (right-edge . -1)
	     (top-edge   . ,(- (get-titlebar-height)))
	     (height     . ,(get-titlebar-height))
	     (background . ,titlebar-background-images))
	    ((class      . title)
	     (left-edge  . ,hpad)
	     (right-edge . ,hpad)
	     (top-edge   . ,stipple-top)
	     (height     . ,(get-stipple-height))
	     (background . ,stipple-images))
	    ((class      . title)
	     (text       . ,window-name)
	     (font       . ,HappyDesktop:font)
	     (x-justify  . center)
	     (y-justify  . center)
	     (top-edge   . ,(+ (- (get-titlebar-height)) 2))
	     (height     . ,font-ht)
	     (width      . ,get-label-width)
	     (left-edge  . ,get-label-left-edge)
	     (background . ,(list col.title-bg-desel col.title-bg))
	     (foreground . ,(list col.title-text-desel col.title-text))) ))

	 (transient-titlebar
	  `(((class      . title)
	     (left-edge  . -1)
	     (right-edge . -1)
	     (top-edge   . ,(- (get-transient-titlebar-height)))
	     (height     . ,(get-transient-titlebar-height))
	     (background . ,transient-titlebar-background-images))
	    ((class      . title)
	     (left-edge  . ,hpad)
	     (right-edge . ,hpad)
	     (top-edge   . ,tr-stipple-top)
	     (height     . ,(get-transient-stipple-height))
	     (background . ,transient-stipple-images)) ))

	 (buttons
	  `(((class . close-button)
	     (removable . t)
	     (left-edge . ,(/ button-wd 2))
	     (top-edge . ,stipple-top)
	     (background . ,close-button-images))
	    ((class . maximize-button)
	     (removable . t)
	     (right-edge . ,(/ button-wd 2))
	     (top-edge . ,stipple-top)
	     (background . ,maximise-button-images))))

	 (transient-buttons
	  `(((class . close-button)
	     (removable . t)
	     (left-edge . ,(/ tr-button-wd 2))
	     (top-edge . ,tr-stipple-top)
	     (background . ,transient-close-button-images))
	    ((class . maximize-button)
	     (removable . t)
	     (right-edge . ,(/ tr-button-wd 2))
	     (top-edge . ,tr-stipple-top)
	     (background . ,transient-maximise-button-images))))

	 (mac-resize-tab
	  `(((class . bottom-right-corner)
	     (removable . t)
	     (bottom-edge . 0)
	     (right-edge . 0)
	     (hidden . ,fixed-size-p)
	     (background . ,right-resize-corner-images))))

	 (x-resize-tab
	  `(((class . bottom-border)
	     (removable . t)
	     (bottom-edge . ,(lambda (w) (- resize-edge-ht)))
	     (right-edge . -1)
	     (left-edge . -1)
	     (hidden . ,fixed-size-p)
	     (height . ,resize-edge-ht)
	     (background . ,bottom-resize-edge-images))
	    ((class . bottom-left-corner)
	     (removable . t)
	     (bottom-edge . ,(lambda (w) (- resize-edge-ht)))
	     (left-edge . 0)
	     (hidden . ,fixed-size-p)
	     (background . ,left-resize-corner-images))
	    ((class . bottom-right-corner)
	     (removable . t)
	     (bottom-edge . ,(lambda (w) (- resize-edge-ht)))
	     (right-edge . 0)
	     (hidden . ,fixed-size-p)
	     (background . ,right-resize-corner-images)) ))

	 (resize-tab
	  (if HappyDesktop:mac-like-resize
	      mac-resize-tab
	      x-resize-tab))

	 (x-shadows
	  `(((class . bottom-border)
	     (removable . t)
	     (right-edge . -1)
	     (left-edge . 0)
	     (height . 1)
	     (background . ,col.borders)
	     (bottom-edge . ,(lambda (w)
			       (if (maximisable-p w)
				   (- -1 resize-edge-ht)
				   -2))) )
	    ((class . right-border)
	     (removable . t)
	     (right-edge . -2)
	     (background . ,col.borders)
	     (width . 1)
	     (top-edge . 0)
	     (bottom-edge . ,(lambda (w)
			       (if (maximisable-p w)
				   (- -1 resize-edge-ht)
				   -2))) )))
	 (mac-shadows
	  `(((class . bottom-border)
	     (removable . t)
	     (right-edge . -1)
	     (left-edge . 0)
	     (height . 1)
	     (background . ,col.borders)
	     (bottom-edge . -2))
	    ((class . right-border)
	     (removable . t)
	     (right-edge . -2)
	     (background . ,col.borders)
	     (width . 1)
	     (top-edge . 0)
	     (bottom-edge . -2))))
	 
	 (page-shadows
	  (if HappyDesktop:shadows
	      (if HappyDesktop:mac-like-resize
		  mac-shadows
		  x-shadows)
	      nil))

	 (plain-shadows
	  (if HappyDesktop:shadows
	      `(((class . right-border)
		 (width . 1)
		 (right-edge . -2)
		 (top-edge . ,(- 1 (get-titlebar-height)))
		 (height . ,(- (get-titlebar-height) 1))
		 (background . ,col.borders)))
	      nil))

	 (transient-shadows
	  (if HappyDesktop:shadows
	      `(((class . right-border)
		 (width . 1)
		 (right-edge . -2)
		 (top-edge . ,(- 1 (get-transient-titlebar-height)))
		 (height . ,(- (get-transient-titlebar-height) 1))
		 (background . ,col.borders)))
	      nil))
	 
	 (borders
	  `(((class . bottom-border)
	     (left-edge . 0)
	     (right-edge . 0)
	     (height . 1)
	     (bottom-edge . -1)
	     (background . ,(list col.borders-desel col.borders)))
	    ((class . right-border)
	     (top-edge . 0)
	     (bottom-edge . -1)
	     (width . 1)
	     (right-edge . -1)
	     (background . ,(list col.borders-desel col.borders)))
	    ((class . left-border)
	     (top-edge . 0)
	     (bottom-edge . -1)
	     (width . 1)
	     (left-edge . -1)
	     (background . ,(list col.borders-desel col.borders))) )) )

    ;; Combine the fragments into frame defs, and store in *-frame:
    (setq shaded-standard-frame (append titlebar buttons)
	  standard-frame (append shaded-standard-frame
				 borders resize-tab
				 page-shadows plain-shadows)
	  shaded-transient-frame (append transient-titlebar transient-buttons)
	  transient-frame (append shaded-transient-frame
				  borders resize-tab
				  page-shadows transient-shadows))
    
    ))


;;;; Theme bootstrapping

(rebuild-colours)
(rebuild-images)
(rebuild-frames)

;; Setup the new frame style.

(add-frame-style
 'HappyDesktop (lambda (w type)
		 (cond ((eq type 'transient)        transient-frame)
		       ((eq type 'shaped-transient) transient-frame)
		       ((eq type 'shaped)           shaded-standard-frame)
		       ((eq type 'shaded)           shaded-standard-frame)
		       ((eq type 'shaded-transient) shaded-transient-frame)
		       ((eq type 'unframed)         nil-frame)
		       (t                           standard-frame))))

;; Rebuild a frame when its title changes.

(call-after-property-changed
 'WM_NAME (lambda (w)
	    (if (eq (window-get w 'current-frame-style) 'HappyDesktop)
		(rebuild-frame w))))
